@@ -34,7 +34,7 @@
  * SUCH DAMAGE.
  */
 
-#include "tftpsubs.h"
+#include "common/tftpsubs.h"
 
 #ifndef lint
 static const char *copyright UNUSED =
@@ -43,7 +43,7 @@ static const char *copyright UNUSED =
 /* static char sccsid[] = "@(#)main.c	8.1 (Berkeley) 6/6/93"; */
 /* static char rcsid[] = "$OpenBSD: main.c,v 1.4 1997/01/17 07:13:30 millert Exp $"; */
 static const char *rcsid UNUSED =
- "tftp-hpa $Id: main.c,v 1.19 2001/11/29 22:02:52 hpa Exp $";
+ "tftp-hpa $Id$";
 #endif /* not lint */
 
 /* Many bug fixes are from Jim Guyton <guyton@rand-unix> */
@@ -183,73 +183,139 @@ char	*tail(char *);
 
 char *xstrdup(const char *);
 
+const char *program;
+
+static inline void usage(int errcode)
+{
+  fprintf(stderr, "Usage: %s [-v][-m mode] [host [port]] [-c command]\n", program);
+  exit(errcode);
+}
+
 int
 main(int argc, char *argv[])
 {
-	struct sockaddr_in s_in;
-	int c;
-	int pargc;
-	char **pargv;
+  struct sockaddr_in s_in;
+  int arg;
+  static int pargc, peerargc;
+  static int iscmd = 0;
+  char **pargv;
+  const char *optx;
+  char *peerargv[3];
 
-	while ((c = getopt(argc, argv, "Vv")) != -1) {
-	  switch (c) {
-	  case 'v':
-	    verbose = 1;
-	    break;
-	  case 'V':
-	    /* Print version and configuration to stdout and exit */
-	    printf("%s\n", TFTP_CONFIG_STR);
-	    exit(0);
-	  default:
-	    fprintf(stderr, "Usage: %s [-v] [host]\n", argv[0]);
-	    exit(EX_USAGE);
+  program = argv[0];
+
+  mode = MODE_DEFAULT;
+
+  peerargv[0] = argv[0];
+  peerargc = 1;
+
+  for ( arg = 1 ; !iscmd && arg < argc ; arg++ ) {
+    if ( argv[arg][0] == '-' ) {
+      for ( optx = &argv[arg][1] ; *optx ; optx++ ) {
+	switch ( *optx ) {
+	case 'v':
+	  verbose = 1;
+	  break;
+	case 'V':
+	  /* Print version and configuration to stdout and exit */
+	  printf("%s\n", TFTP_CONFIG_STR);
+	  exit(0);
+	case 'm':
+	  if ( ++arg >= argc )
+	    usage(EX_USAGE);
+	  {
+	    const struct modes *p;
+	    
+	    for ( p = modes ; p->m_name ; p++ ) {
+	      if (!strcmp(argv[arg], p->m_name))
+		break;
+	    }
+	    if (p->m_name) {
+	      settftpmode(p);
+	    } else {
+	      fprintf(stderr, "%s: invalid mode: %s\n", argv[0], argv[arg]);
+	      exit(EX_USAGE);
+	    }
 	  }
+	  break;
+	case 'c':
+	  iscmd = 1;
+	  break;
+	case 'h':
+	default:
+	  usage(*optx == 'h' ? 0 : EX_USAGE);
 	}
+      }
+    } else {
+      if ( peerargc >= 3 )
+	usage(EX_USAGE);
 
-	pargc = argc - (optind-1);
-	pargv = argv + (optind-1);
+      peerargv[peerargc++] = argv[arg];
+    }
+  }
+  
+  pargv = argv + arg;
+  pargc = argc - arg;
+  
+  sp = getservbyname("tftp", "udp");
+  if (sp == 0) {
+    /* Use canned values */
+    if (verbose)
+      fprintf(stderr, "tftp: tftp/udp: unknown service, faking it...\n");
+    sp = xmalloc(sizeof(struct servent));
+    sp->s_name    = (char *)"tftp";
+    sp->s_aliases = NULL;
+    sp->s_port    = htons(IPPORT_TFTP);
+    sp->s_proto   = (char *)"udp";
+  }
+  port = sp->s_port;		/* Default port */
+  f = socket(AF_INET, SOCK_DGRAM, 0);
+  if (f < 0) {
+    perror("tftp: socket");
+    exit(EX_OSERR);
+  }
+  bzero((char *)&s_in, sizeof (s_in));
+  s_in.sin_family = AF_INET;
+  if (bind(f, (struct sockaddr *)&s_in, sizeof (s_in)) < 0) {
+    perror("tftp: bind");
+    exit(EX_OSERR);
+  }
+  bsd_signal(SIGINT, intr);
 
-	sp = getservbyname("tftp", "udp");
-	if (sp == 0) {
-		/* Use canned values */
-		fprintf(stderr, "tftp: tftp/udp: unknown service, faking it...\n");
-		sp = xmalloc(sizeof(struct servent));
-		sp->s_name    = (char *)"tftp";
-		sp->s_aliases = NULL;
-		sp->s_port    = htons(IPPORT_TFTP);
-		sp->s_proto   = (char *)"udp";
-		exit(1);
-	}
-	f = socket(AF_INET, SOCK_DGRAM, 0);
-	if (f < 0) {
-		perror("tftp: socket");
-		exit(3);
-	}
-	bzero((char *)&s_in, sizeof (s_in));
-	s_in.sin_family = AF_INET;
-	if (bind(f, (struct sockaddr *)&s_in, sizeof (s_in)) < 0) {
-		perror("tftp: bind");
-		exit(1);
-	}
-	mode = MODE_DEFAULT;
-	bsd_signal(SIGINT, intr);
-	if (pargc > 1) {
-		if (sigsetjmp(toplevel,1) != 0)
-			exit(0);
-		setpeer(pargc, pargv);
-	}
-	if (sigsetjmp(toplevel,1) != 0)
-		(void)putchar('\n');
+  if ( peerargc ) {
+    /* Set peer */
+    if (sigsetjmp(toplevel,1) != 0)
+      exit(EX_NOHOST);
+    setpeer(peerargc, peerargv);
+  }
 
+  if ( iscmd && pargc ) {
+    /* -c specified; execute command and exit */
+    struct cmd *c;
+    
+    if (sigsetjmp(toplevel,1) != 0)
+      exit(EX_UNAVAILABLE);
+    
+    c = getcmd(pargv[0]);
+    if ( c == (struct cmd *)-1 || c == (struct cmd *)0 ) {
+      fprintf(stderr, "%s: invalid command: %s\n", argv[0], pargv[1]);
+      exit(EX_USAGE);
+    }	
+    (*c->handler)(pargc, pargv);
+    exit(0);
+  }
+  
 #ifdef WITH_READLINE
 #ifdef HAVE_READLINE_HISTORY_H
-	using_history();
+  using_history();
 #endif
 #endif
-
-	command();
-
-	return 0;		/* Never reached */
+  
+  if (sigsetjmp(toplevel,1) != 0)
+    (void)putchar('\n');
+  command();
+  
+  return 0;		/* Never reached */
 }
 
 char    *hostname;
@@ -270,8 +336,10 @@ getmoreargs(const char *partial, const char *mprompt)
   
   elen = strlen(eline);
 
-  if (line)
+  if (line) {
 	  free(line);
+	  line = NULL;
+  }
   line = xmalloc(len+elen+1);
   strcpy(line, partial);
   strcpy(line+len, eline);
@@ -438,7 +506,7 @@ put(int argc, char *argv[])
 			herror((char *)NULL);
 			return;
 		}
-		bcopy(hp->h_addr, (caddr_t)&peeraddr.sin_addr, hp->h_length);
+		bcopy(hp->h_addr, &peeraddr.sin_addr, hp->h_length);
 		peeraddr.sin_family = hp->h_addrtype;
 		connected = 1;
 		hostname = xstrdup(hp->h_name);
@@ -668,8 +736,10 @@ command(void)
 
 	for (;;) {
 #ifdef WITH_READLINE
-	        if ( line )
+	        if ( line ) {
 			free(line);
+			line = NULL;
+		}
 	        line = readline(prompt);
 		if ( !line )
 			exit(0); /* EOF */
